/*
 *	psymodel.c
 *
 *	Copyright (c) 1999 Mark Taylor
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


/*
PSYCHO ACOUSTICS


This routine computes the psycho acoustics, delayed by
one granule.  

Input: buffer of PCM data (1024 samples).  

This window should be centered over the 576 sample granule window.
The routine will compute the psycho acoustics for
this granule, but return the psycho acoustics computed
for the *previous* granule.  This is because the block
type of the previous granule can only be determined
after we have computed the psycho acoustics for the following
granule.  

Output:  maskings and energies for each scalefactor band.
block type, PE, and some correlation measures.  
The PE is used by CBR modes to determine if extra bits
from the bit reservoir should be used.  The correlation
measures are used to determine mid/side or regular stereo.


Notation:

barks:  a non-linear frequency scale.  Mapping from frequency to
        barks is given by freq2bark()

scalefactor bands: The spectrum (frequencies) are broken into 
                   SBMAX "scalefactor bands".  Thes bands
                   are determined by the MPEG ISO spec.  In
                   the noise shaping/quantization code, we allocate
                   bits among the partition bands to achieve the
                   best possible quality

partition bands:   The spectrum is also broken into about
                   64 "partition bands".  Each partition 
                   band is about .34 barks wide.  There are about 2-5
                   partition bands for each scalefactor band.

LAME computes all psycho acoustic information for each partition
band.  Then at the end of the computations, this information
is mapped to scalefactor bands.  The energy in each scalefactor
band is taken as the sum of the energy in all partition bands
which overlap the scalefactor band.  The maskings can be computed
in the same way (and thus represent the average masking in that band)
or by taking the minmum value multiplied by the number of
partition bands used (which represents a minimum masking in that band).


The general outline is as follows:


1. compute the energy in each partition band
2. compute the tonality in each partition band
3. compute the strength of each partion band "masker"
4. compute the masking (via the spreading function applied to each masker)
5. Modifications for mid/side masking.  

Each partition band is considiered a "masker".  The strength
of the i'th masker in band j is given by:

    s3(bark(i)-bark(j))*strength(i)

The strength of the masker is a function of the energy and tonality.
The more tonal, the less masking.  LAME uses a simple linear formula
(controlled by NMT and TMN) which says the strength is given by the
energy divided by a linear function of the tonality.


s3() is the "spreading function".  It is given by a formula
determined via listening tests.  

The total masking in the j'th partition band is the sum over
all maskings i.  It is thus given by the convolution of
the strength with s3(), the "spreading function."

masking(j) = sum_over_i  s3(i-j)*strength(i)  = s3 o strength

where "o" = convolution operator.  s3 is given by a formula determined
via listening tests.  It is normalized so that s3 o 1 = 1.

Note: instead of a simple convolution, LAME also has the
option of using "additive masking"

The most critical part is step 2, computing the tonality of each
partition band.  LAME has two tonality estimators.  The first
is based on the ISO spec, and measures how predictiable the
signal is over time.  The more predictable, the more tonal.
The second measure is based on looking at the spectrum of
a single granule.  The more peaky the spectrum, the more
tonal.  By most indications, the latter approach is better.

Finally, in step 5, the maskings for the mid and side
channel are possibly increased.  Under certain circumstances,
noise in the mid & side channels is assumed to also
be masked by strong maskers in the L or R channels.


Other data computed by the psy-model:

ms_ratio        side-channel / mid-channel masking ratio (for previous granule)
ms_ratio_next   side-channel / mid-channel masking ratio for this granule

percep_entropy[2]     L and R values (prev granule) of PE - A measure of how 
                      much pre-echo is in the previous granule
percep_entropy_MS[2]  mid and side channel values (prev granule) of percep_entropy
energy[4]             L,R,M,S energy in each channel, prev granule
blocktype_d[2]        block type to use for previous granule

*/

#include <assert.h>

#include "config.h"
#include "global.h"

#include "util.h"
#include "psymodel.h"
#include "tables.h"
#include "lame.h"

#define rpelev 2
#define rpelev2 16

/* size of each partition band, in barks: */
#define DELBARK .34

#if 1
    /* AAC values, results in more masking over MP3 values */
# define TMN 18
# define NMT 6
#else
    /* MP3 values */
# define TMN 29
# define NMT 6
#endif

#define NBPSY_l  (SBMAX_l)
#define NBPSY_s  (SBMAX_s)


#ifdef M_LN10
#define		LN_TO_LOG10		(M_LN10/10)
#else
#define LN_TO_LOG10 0.2302585093
#endif
#define LOG10 2.30258509299404568402

static void
ms_convert_long(FLOAT src[2][BLKSIZE])
{
	int	j;

	for (j = BLKSIZE-1; j >=0 ; --j) {
		FLOAT l = src[0][j];
		FLOAT r = src[1][j];
		src[0][j] = (l+r)*(FLOAT)(SQRT2*0.5);
		src[1][j] = (l-r)*(FLOAT)(SQRT2*0.5);
	}
}

static void
ms_convert_short(FLOAT src[2][3][BLKSIZE_s])
{
#if 0
	int	b,j;

	for (b = 2; b >= 0; --b) {
		for (j = BLKSIZE_s-1; j >= 0 ; --j) {
			FLOAT l = src[0][b][j];
			FLOAT r = src[1][b][j];
			src[0][b][j] = (l+r)*(FLOAT)(SQRT2*0.5);
			src[1][b][j] = (l-r)*(FLOAT)(SQRT2*0.5);
		}
	}
#else
	int	j;

	for (j = BLKSIZE_s*3-1; j >= 0 ; --j) {
		FLOAT l = src[0][0][j];
		FLOAT r = src[1][0][j];
		src[0][0][j] = (l+r)*(FLOAT)(SQRT2*0.5);
		src[1][0][j] = (l-r)*(FLOAT)(SQRT2*0.5);
	}
#endif
}


/*
   L3psycho_anal.  Compute psycho acoustics.

   Data returned to the calling program must be delayed by one 
   granule. 

   This is done in two places.  
   If we do not need to know the blocktype, the copying
   can be done here at the top of the program: we copy the data for
   the last granule (computed during the last call) before it is
   overwritten with the new data.  It looks like this:
  
   0. static psymodel_data 
   1. calling_program_data = psymodel_data
   2. compute psymodel_data
    
   For data which needs to know the blocktype, the copying must be
   done at the end of this loop, and the old values must be saved:
   
   0. static psymodel_data_old 
   1. compute psymodel_data
   2. compute possible block type of this granule
   3. compute final block type of previous granule based on #2.
   4. calling_program_data = psymodel_data_old
   5. psymodel_data_old = psymodel_data
*/
int L3psycho_anal(gogo_thread_data *tl,
                    /* const sample_t *buffer[2], */ int gr_out, 
                    FLOAT8 *ms_ratio,
                    FLOAT8 *ms_ratio_next,
		    III_psy_ratio masking_ratio[2][2],
		    III_psy_ratio masking_MS_ratio[2][2],
		    FLOAT8 percep_entropy[2],FLOAT8 percep_MS_entropy[2], 
                    FLOAT8 energy[4],
                    int blocktype_d[2])
{
/* to get a good cache performance, one has to think about
 * the sequence, in which the variables are used.  
 * (Note: these static variables have been moved to the RW. struct,
 * and their order in memory is layed out in util.h)
 */

  /* fft and energy calculation   */
  FLOAT (*wsamp_l)[BLKSIZE];
  FLOAT (*wsamp_s)[3][BLKSIZE_s];

  /* convolution   */
  FLOAT8 eb[CBANDS];
  FLOAT8 cb[CBANDS];
  FLOAT8 thr[CBANDS];
  
  /* ratios    */
  FLOAT8 ms_ratio_l=0,ms_ratio_s=0;

  /* block type  */
  int blocktype[2],uselongblock[2];
  
  /* usual variables like loop indices, etc..    */
  int numchn, chn;
  int   b, i, j, k;
  int	sb,sblock;


  if(RO.psymodel_init==0) {
      psymodel_init();
      RO.psymodel_init=1;
  }
  
  numchn = RO.channels_out;
  /* chn=2 and 3 = Mid and Side channels */
  if (RO.mode == JOINT_STEREO) numchn=4;

  for (chn=0; chn<numchn; chn++) {
	  energy[chn]=RW.tot_ener[chn];
  }
  
  for (chn=0; chn<numchn; chn++) {

      /* there is a one granule delay.  Copy maskings computed last call
       * into masking_ratio to return to calling program.
       */
      if (chn < 2) {    
	  /* LR maskings  */
	  percep_entropy            [chn]       = RW.pe  [chn]; 
	  masking_ratio    [gr_out] [chn]  .en  = RW.en  [chn];
	  masking_ratio    [gr_out] [chn]  .thm = RW.thm [chn];
      } else {
	  /* MS maskings  */
	  percep_MS_entropy         [chn-2]     = RW.pe  [chn]; 
	  masking_MS_ratio [gr_out] [chn-2].en  = RW.en  [chn];
	  masking_MS_ratio [gr_out] [chn-2].thm = RW.thm [chn];
      }
      
    /* FFT data for mid and side channel is derived from L & R */
    if (chn == 2) {
        ms_convert_long (tl->wsamp_L[gr_out]);
        ms_convert_short(tl->wsamp_S[gr_out]);
    }


    /**********************************************************************
     *  compute energies
     **********************************************************************/
    
//clkbegin();
/* 3.4% */
		wsamp_s = &tl->wsamp_S[gr_out][chn & 1];
		wsamp_l = &tl->wsamp_L[gr_out][chn & 1];

		tl->energy[0]  = (*wsamp_l)[0];
		tl->energy[0] *= tl->energy[0];
    
		RW.tot_ener[chn] = tl->energy[0]; /* sum total energy at nearly no extra cost */
    
		for (j=BLKSIZE/2-1; j >= 0; --j) {
			FLOAT re = (*wsamp_l)[BLKSIZE/2-j];	/*    1,   2,,,,511,512 */
			FLOAT im = (*wsamp_l)[BLKSIZE/2+j];	/* 1023,1022,,,,513,512 */
			tl->energy[BLKSIZE/2-j] = (re * re + im * im) * (FLOAT)0.5;

			if (BLKSIZE/2-j > 10)	/* 11,12,,,,511,512 */
				RW.tot_ener[chn] += tl->energy[BLKSIZE/2-j];
		}
		for (b = 2; b >= 0; --b) {
			tl->energy_s[b][0]  = (*wsamp_s)[b][0];
			tl->energy_s[b][0] *=  tl->energy_s[b][0];
			for (j=BLKSIZE_s/2-1; j >= 0; --j) {
        			FLOAT re = (*wsamp_s)[b][BLKSIZE_s/2-j];
        			FLOAT im = (*wsamp_s)[b][BLKSIZE_s/2+j];
				tl->energy_s[b][BLKSIZE_s/2-j] = (re * re + im * im) * (FLOAT)0.5;
			}
		}
//clkend();

    /**********************************************************************
     *    compute unpredicatability of first six spectral lines            * 
     **********************************************************************/
    for ( j = 0; j < RO.cw_lower_index; j++ )
      {	 /* calculate unpredictability measure cw */
	FLOAT an, a1, a2;
	FLOAT bn, b1, b2;
	FLOAT rn, r1, r2;
	FLOAT numre, numim, den;

	a2 = RW. ax_sav[chn][1][j];
	b2 = RW. bx_sav[chn][1][j];
	r2 = RW. rx_sav[chn][1][j];
	a1 = RW. ax_sav[chn][1][j] = RW. ax_sav[chn][0][j];
	b1 = RW. bx_sav[chn][1][j] = RW. bx_sav[chn][0][j];
	r1 = RW. rx_sav[chn][1][j] = RW. rx_sav[chn][0][j];
	an = RW. ax_sav[chn][0][j] = (*wsamp_l)[j];
	bn = RW. bx_sav[chn][0][j] = j==0 ? (*wsamp_l)[0] : (*wsamp_l)[BLKSIZE-j];  
	rn = RW. rx_sav[chn][0][j] = sqrt(tl->energy[j]);

	{ /* square (x1,y1) */
	  if( r1 != 0 ) {
	    numre = (a1*b1);
	    numim = (a1*a1-b1*b1)*(FLOAT)0.5;
	    den = r1*r1;
	  } else {
	    numre = 1;
	    numim = 0;
	    den = 1;
	  }
	}
	
	{ /* multiply by (x2,-y2) */
	  if( r2 != 0 ) {
	    FLOAT tmp2 = (numim+numre)*(a2+b2)*(FLOAT)0.5;
	    FLOAT tmp1 = -a2*numre+tmp2;
	    numre =       -b2*numim+tmp2;
	    numim = tmp1;
	    den *= r2;
	  } else {
	    /* do nothing */
	  }
	}
	
	{ /* r-prime factor */
	  FLOAT tmp = (2*r1-r2)/den;
	  numre *= tmp;
	  numim *= tmp;
	}
	den=rn+fabs(2*r1-r2);
	if( den != 0 ) {
	  numre = (an+bn)*(FLOAT)0.5-numre;
	  numim = (an-bn)*(FLOAT)0.5-numim;
	  den = sqrt(numre*numre+numim*numim)/den;
	}
	RW.cw[j] = den;
      }



    /**********************************************************************
     *     compute unpredicatibility of next 200 spectral lines            *
     **********************************************************************/ 
//clkbegin();
/* 5.4% */
    for ( j = RO.cw_lower_index; j < RO.cw_upper_index; j += 4 )
      {/* calculate unpredictability measure cw */
	FLOAT rn, r1, r2;
	FLOAT numre, numim, den;
	
	k = (j+2) / 4; 
	
	{ /* square (x1,y1) */
	  r1 = tl->energy_s[0][k];
	  if( r1 != 0 ) {
	    FLOAT a1 = (*wsamp_s)[0][k]; 
	    FLOAT b1 = (*wsamp_s)[0][BLKSIZE_s-k]; /* k is never 0 */
	    numre = (a1*b1);
	    numim = (a1*a1-b1*b1)*(FLOAT)0.5;
	    den = r1;
	    r1 = sqrt(r1);
	  } else {
	    numre = 1;
	    numim = 0;
	    den = 1;
	  }
	}
	
	
	{ /* multiply by (x2,-y2) */
	  r2 = tl->energy_s[2][k];
	  if( r2 != 0 ) {
	    FLOAT a2 = (*wsamp_s)[2][k]; 
	    FLOAT b2 = (*wsamp_s)[2][BLKSIZE_s-k];
	    
	    
	    FLOAT tmp2 = (numim+numre)*(a2+b2)*(FLOAT)0.5;
	    FLOAT tmp1 = -a2*numre+tmp2;
	    numre =       -b2*numim+tmp2;
	    numim = tmp1;
	    
	    r2 = sqrt(r2);
	    den *= r2;
	  } else {
	    /* do nothing */
	  }
	}
	
	{ /* r-prime factor */
	  FLOAT tmp = (2*r1-r2)/den;
	  numre *= tmp;
	  numim *= tmp;
	}
	
	rn = sqrt(tl->energy_s[1][k]);
	den=rn+fabs(2*r1-r2);
	if( den != 0 ) {
	  FLOAT an = (*wsamp_s)[1][k]; 
	  FLOAT bn = (*wsamp_s)[1][BLKSIZE_s-k];
	  numre = (an+bn)*(FLOAT)0.5-numre;
	  numim = (an-bn)*(FLOAT)0.5-numim;
	  den = sqrt(numre*numre+numim*numim)/den;
	}
	
	RW.cw[j+1] = RW.cw[j+2] = RW.cw[j+3] = RW.cw[j] = den;
      }
//clkend();
    
#if 0
    for ( j = 14; j < HBLKSIZE-4; j += 4 )
      {/* calculate energy from short ffts */
	FLOAT8 tot,ave;
	k = (j+2) / 4; 
	for (tot=0, sblock=0; sblock < 3; sblock++)
	  tot+=tl->energy_s[sblock][k];
	ave = tl->energy[j+1]+ tl->energy[j+2]+ tl->energy[j+3]+ tl->energy[j];
	ave /= 4.;
	tl->energy[j+1] = tl->energy[j+2] = tl->energy[j+3] =  tl->energy[j]=tot;
      }
#endif
    
    /**********************************************************************
     *    Calculate the energy and the unpredictability in the threshold   *
     *    calculation partitions                                           *
     **********************************************************************/

    b = 0;
    for (j = 0; j < RO.cw_upper_index && RO.numlines_l[b] && b<RO.npart_l_orig; )
      {
	FLOAT8 ebb, cbb;

	ebb = tl->energy[j];
	cbb = tl->energy[j] * RW.cw[j];
	j++;

	for (i = RO.numlines_l[b] - 1; i > 0; i--)
	  {
	    ebb += tl->energy[j];
	    cbb += tl->energy[j] * RW.cw[j];
	    j++;
	  }
	eb[b] = ebb;
	cb[b] = cbb;
	b++;
      }

    for (; b < RO.npart_l_orig; b++ )
      {
	FLOAT8 ebb = tl->energy[j++];
	assert(RO.numlines_l[b]);
	for (i = RO.numlines_l[b] - 1; i > 0; i--)
	  {
	    ebb += tl->energy[j++];
	  }
	eb[b] = ebb;
	cb[b] = ebb * 0.4;
      }

    /**********************************************************************
     *      convolve the partitioned energy and unpredictability           *
     *      with the spreading function, s3_l[b][k]                        *
     ******************************************************************** */
//clkbegin();
/* 15.3% */
    RW.pe[chn] = 0;		/*  calculate percetual entropy */
    for ( b = 0;b < RO.npart_l; b++ )
      {
	FLOAT8 tbb,ecb,ctb;

	ecb = 0;
	ctb = 0;

	for ( k = RO.s3ind[b][0]; k <= RO.s3ind[b][1]; k++ )
	  {
	    ecb += RO.s3_l[b][k] * eb[k];	/* sprdngf for Layer III */
	    ctb += RO.s3_l[b][k] * cb[k];
	  }

	/* calculate the tonality of each threshold calculation partition 
	 * calculate the SNR in each threshold calculation partition 
	 * tonality = -0.299 - .43*log(ctb/ecb);
	 * tonality = 0:           use NMT   (lots of masking)
	 * tonality = 1:           use TMN   (little masking)
	 */

/* ISO values */
#define CONV1 (-.299)
#define CONV2 (-.43)

	tbb = ecb;
	if (tbb != 0)
	  {
	    tbb = ctb / tbb;
	    if (tbb <= exp((1-CONV1)/CONV2)) 
	      { /* tonality near 1 */
		tbb = exp(-LN_TO_LOG10 * TMN);
	      }
	    else if (tbb >= exp((0-CONV1)/CONV2)) 
	      {/* tonality near 0 */
		tbb = exp(-LN_TO_LOG10 * NMT);
	      }
	    else
	      {
		/* convert to tonality index */
		/* tonality small:   tbb=1 */
		/* tonality large:   tbb=-.299 */
		tbb = CONV1 + CONV2*log(tbb);
		tbb = exp(-LN_TO_LOG10 * ( TMN*tbb + (1-tbb)*NMT) );
	      }
	  }

	/* at this point, tbb represents the amount the spreading function
	 * will be reduced.  The smaller the value, the less masking.
	 * minval[] = 1 (0db)     says just use tbb.
	 * minval[]= .01 (-20db)  says reduce spreading function by 
	 *                        at least 20db.  
	 */

	tbb = Min(RO.minval[b], tbb);
	ecb *= tbb;

	/* long block pre-echo control.   */
	/* rpelev=2.0, rpelev2=16.0 */
	/* note: all surges in PE are because of this pre-echo formula
	 * for thr[b].  If it this is not used, PE is always around 600
	 */
	/* dont use long block pre-echo control if previous granule was 
	 * a short block.  This is to avoid the situation:   
	 * frame0:  quiet (very low masking)  
	 * frame1:  surge  (triggers short blocks)
	 * frame2:  regular frame.  looks like pre-echo when compared to 
	 *          frame0, but all pre-echo was in frame1.
	 */
	/* chn=0,1   L and R channels
	   chn=2,3   S and M channels.  
	*/
        
        if (RW.blocktype_old[chn>1 ? chn-2 : chn] == SHORT_TYPE )
	  thr[b] = ecb; /* Min(ecb, rpelev*RW.nb_1[chn][b]); */
	else
	  thr[b] = Min(ecb, Min(rpelev*RW.nb_1[chn][b],rpelev2*RW.nb_2[chn][b]) );

	RW.nb_2[chn][b] = RW.nb_1[chn][b];
	RW.nb_1[chn][b] = ecb;

	{
	  FLOAT8 thrpe;
	  thrpe = Max(thr[b],RO.ATH.cb[b]);
	  /*
	    printf("%i thr=%e   ATH=%e  \n",b,thr[b],RO.ATH.cb[b]);
	  */
	  if (thrpe < eb[b])
	    RW.pe[chn] -= RO.numlines_l[b] * log(thrpe / eb[b]);
	}
      }
//clkend();

    /*************************************************************** 
     * determine the block type (window type) based on L & R channels
     * 
     ***************************************************************/
    {  /* compute PE for all 4 channels */
      FLOAT mn,mx,ma=0,mb=0,mc=0;
      
      for ( j = HBLKSIZE_s/2; j < HBLKSIZE_s; j ++)
	{
	  ma += tl->energy_s[0][j];
	  mb += tl->energy_s[1][j];
	  mc += tl->energy_s[2][j];
	}
      mn = Min(ma,mb);
      mn = Min(mn,mc);
      mx = Max(ma,mb);
      mx = Max(mx,mc);
      
      /* bit allocation is based on pe.  */
      if (mx>mn) {
	FLOAT8 tmp = 400*log(mx/(1e-12+mn));
	if (tmp>RW.pe[chn]) RW.pe[chn]=tmp;
      }

      /* block type is based just on L or R channel */      
      if (chn<2) {
	uselongblock[chn] = 1;
	
	/* tuned for t1.wav.  doesnt effect most other samples */
	if (RW.pe[chn] > 3000) 
	  uselongblock[chn]=0;
	
	if ( mx > 30*mn ) 
	  {/* big surge of energy - always use short blocks */
	    uselongblock[chn] = 0;
	  } 
	else if ((mx > 10*mn) && (RW.pe[chn] > 1000))
	  {/* medium surge, medium pe - use short blocks */
	    uselongblock[chn] = 0;
	  }
	
	/* disable short blocks */
      }
    }

    /*************************************************************** 
     * compute masking thresholds for both short and long blocks
     ***************************************************************/
    /* longblock threshold calculation (part 2) */
    for ( sb = 0; sb < NBPSY_l; sb++ )
      {
	FLOAT8 enn = RO.w1_l[sb] * eb[RO.bu_l[sb]] + RO.w2_l[sb] * eb[RO.bo_l[sb]];
	FLOAT8 thmm = RO.w1_l[sb] *thr[RO.bu_l[sb]] + RO.w2_l[sb] * thr[RO.bo_l[sb]];

        for ( b = RO.bu_l[sb]+1; b < RO.bo_l[sb]; b++ )
          {
            enn  += eb[b];
            thmm += thr[b];
          }

	RW.en [chn].l[sb] = enn;
	RW.thm[chn].l[sb] = thmm;
      }
    

    /* threshold calculation for short blocks */
    for ( sblock = 0; sblock < 3; sblock++ )
      {
	j = 0;
	for ( b = 0; b < RO.npart_s_orig; b++ )
	  {
	    FLOAT ecb = tl->energy_s[sblock][j++];
	    for (i = 1 ; i<RO.numlines_s[b]; ++i)
	      {
		ecb += tl->energy_s[sblock][j++];
	      }
	    eb[b] = ecb;
	  }

	for ( b = 0; b < RO.npart_s; b++ )
	  {
	    FLOAT8 ecb = 0;
	    for ( k = RO.s3ind_s[b][0]; k <= RO.s3ind_s[b][1]; k++ ) {
		ecb += RO.s3_s[b][k] * eb[k];
	    }
            ecb *= RO.SNR_s[b];
	    thr[b] = Max (1e-6, ecb);
	  }

	for ( sb = 0; sb < NBPSY_s; sb++ ){
            FLOAT8 enn  = RO.w1_s[sb] * eb[RO.bu_s[sb]] + RO.w2_s[sb] * eb[RO.bo_s[sb]];
	    FLOAT8 thmm = RO.w1_s[sb] *thr[RO.bu_s[sb]] + RO.w2_s[sb] * thr[RO.bo_s[sb]];
	    
            for ( b = RO.bu_s[sb]+1; b < RO.bo_s[sb]; b++ ) {
		enn  += eb[b];
		thmm += thr[b];
	    }

	    RW.en [chn].s[sb][sblock] = enn;
	    RW.thm[chn].s[sb][sblock] = thmm;
	  }
      }
  } /* end loop over chn */


  /* compute M/S thresholds from Johnston & Ferreira 1992 ICASSP paper */
  if (RO.mode == JOINT_STEREO) {
    FLOAT8 rside,rmid,mld;
    int chmid=2,chside=3; 
    
    for ( sb = 0; sb < NBPSY_l; sb++ ) {
      /* use this fix if L & R masking differs by 2db or less */
      /* if db = 10*log10(x2/x1) < 2 */
      /* if (x2 < 1.58*x1) { */
      if (RW.thm[0].l[sb] <= 1.58*RW.thm[1].l[sb]
	  && RW.thm[1].l[sb] <= 1.58*RW.thm[0].l[sb]) {

	mld = RO.mld_l[sb]*RW.en[chside].l[sb];
	rmid = Max(RW.thm[chmid].l[sb], Min(RW.thm[chside].l[sb],mld));

	mld = RO.mld_l[sb]*RW.en[chmid].l[sb];
	rside = Max(RW.thm[chside].l[sb],Min(RW.thm[chmid].l[sb],mld));

	RW.thm[chmid].l[sb]=rmid;
	RW.thm[chside].l[sb]=rside;
      }
    }
    for ( sb = 0; sb < NBPSY_s; sb++ ) {
      for ( sblock = 0; sblock < 3; sblock++ ) {
	if (RW.thm[0].s[sb][sblock] <= 1.58*RW.thm[1].s[sb][sblock]
	    && RW.thm[1].s[sb][sblock] <= 1.58*RW.thm[0].s[sb][sblock]) {

	  mld = RO.mld_s[sb]*RW.en[chside].s[sb][sblock];
	  rmid = Max(RW.thm[chmid].s[sb][sblock],Min(RW.thm[chside].s[sb][sblock],mld));

	  mld = RO.mld_s[sb]*RW.en[chmid].s[sb][sblock];
	  rside = Max(RW.thm[chside].s[sb][sblock],Min(RW.thm[chmid].s[sb][sblock],mld));

	  RW.thm[chmid].s[sb][sblock]=rmid;
	  RW.thm[chside].s[sb][sblock]=rside;
	}
      }
    }
  }

  if (RO.mode == JOINT_STEREO)  {
    /* determin ms_ratio from masking thresholds*/
    /* use ms_stereo (ms_ratio < .35) if average thresh. diff < 5 db */
    FLOAT8 db,x1,x2,sidetot=0,tot=0;
    for (sb= NBPSY_l/4 ; sb< NBPSY_l; sb ++ ) {
      x1 = Min(RW.thm[0].l[sb],RW.thm[1].l[sb]);
      x2 = Max(RW.thm[0].l[sb],RW.thm[1].l[sb]);
      /* thresholds difference in db */
      if (x2 >= 1000*x1)  db=3;
      else db = log10(x2/x1);  
      /*  RO.printf("db = %f %e %e  \n",db,RW.thm[0].l[sb],RW.thm[1].l[sb]);*/
      sidetot += db;
      tot++;
    }
    ms_ratio_l= (sidetot/tot)*0.7; /* was .35*(sidetot/tot)/5.0*10 */
    ms_ratio_l = Min(ms_ratio_l,0.5);
    
    sidetot=0; tot=0;
    for ( sblock = 0; sblock < 3; sblock++ )
      for ( sb = NBPSY_s/4; sb < NBPSY_s; sb++ ) {
	x1 = Min(RW.thm[0].s[sb][sblock],RW.thm[1].s[sb][sblock]);
	x2 = Max(RW.thm[0].s[sb][sblock],RW.thm[1].s[sb][sblock]);
	/* thresholds difference in db */
	if (x2 >= 1000*x1)  db=3;
	else db = log10(x2/x1);  
	sidetot += db;
	tot++;
      }
    ms_ratio_s = (sidetot/tot)*0.7; /* was .35*(sidetot/tot)/5.0*10 */
    ms_ratio_s = Min(ms_ratio_s,.5);
  }

  /*************************************************************** 
   * determine final block type
   ***************************************************************/

  for (chn=0; chn<RO.channels_out; chn++) {
    blocktype[chn] = NORM_TYPE;
  }

  if (RO.channels_out==2) {
    if (RO.mode==JOINT_STEREO) {
      /* force both channels to use the same block type */
      /* this is necessary if the frame is to be encoded in ms_stereo.  */
      /* But even without ms_stereo, FhG  does this */
      int bothlong= (uselongblock[0] && uselongblock[1]);
      if (!bothlong) {
	uselongblock[0]=0;
	uselongblock[1]=0;
      }
    }
  }

  
  
  /* update the blocktype of the previous granule, since it depends on what
   * happend in this granule */
  for (chn=0; chn<RO.channels_out; chn++) {
    if ( uselongblock[chn])
      {				/* no attack : use long blocks */
	assert( RW.blocktype_old[chn] != START_TYPE );
	switch( RW.blocktype_old[chn] ) 
	  {
	  case NORM_TYPE:
	  case STOP_TYPE:
	    blocktype[chn] = NORM_TYPE;
	    break;
	  case SHORT_TYPE:
	    blocktype[chn] = STOP_TYPE; 
	    break;
	  }
      } else   {
	/* attack : use short blocks */
	blocktype[chn] = SHORT_TYPE;
	if ( RW.blocktype_old[chn] == NORM_TYPE ) {
	  RW.blocktype_old[chn] = START_TYPE;
	}
	if ( RW.blocktype_old[chn] == STOP_TYPE ) {
	  RW.blocktype_old[chn] = SHORT_TYPE ;
	}
      }
    
    blocktype_d[chn] = RW.blocktype_old[chn];  /* value returned to calling program */
    RW.blocktype_old[chn] = blocktype[chn];    /* save for next call to l3psy_anal */
  }
  
  if (blocktype_d[0]==2) 
    *ms_ratio = RW.ms_ratio_s_old;
  else
    *ms_ratio = RW.ms_ratio_l_old;

  RW.ms_ratio_s_old = ms_ratio_s;
  RW.ms_ratio_l_old = ms_ratio_l;

  /* we dont know the block type of this frame yet - assume long */
  *ms_ratio_next = ms_ratio_l;

  return 0;
}

/* addition of simultaneous masking   Naoki Shibata 2000/7 */
INLINE static FLOAT8 mask_add(FLOAT8 m1,FLOAT8 m2,int k,int b, gogo_thread_data *tl)
{
  static const FLOAT8 table1[] = {
    3.3246 *3.3246 ,3.23837*3.23837,3.15437*3.15437,3.00412*3.00412,2.86103*2.86103,2.65407*2.65407,2.46209*2.46209,2.284  *2.284  ,
    2.11879*2.11879,1.96552*1.96552,1.82335*1.82335,1.69146*1.69146,1.56911*1.56911,1.46658*1.46658,1.37074*1.37074,1.31036*1.31036,
    1.25264*1.25264,1.20648*1.20648,1.16203*1.16203,1.12765*1.12765,1.09428*1.09428,1.0659 *1.0659 ,1.03826*1.03826,1.01895*1.01895,
    1
  };

  static const FLOAT8 table2[] = {
    1.33352*1.33352,1.35879*1.35879,1.38454*1.38454,1.39497*1.39497,1.40548*1.40548,1.3537 *1.3537 ,1.30382*1.30382,1.22321*1.22321,
    1.14758*1.14758
  };

  static const FLOAT8 table3[] = {
    2.35364*2.35364,2.29259*2.29259,2.23313*2.23313,2.12675*2.12675,2.02545*2.02545,1.87894*1.87894,1.74303*1.74303,1.61695*1.61695,
    1.49999*1.49999,1.39148*1.39148,1.29083*1.29083,1.19746*1.19746,1.11084*1.11084,1.03826*1.03826
  };


  int i;
  FLOAT8 m;

  if (m1 == 0) return m2;

  if (b < 0) b = -b;

  i = 10*log10(m2 / m1)/10*16;
  m = 10*log10((m1+m2)/RO.ATH.cb[k]);

  if (i < 0) i = -i;

  if (b <= 3) {  /* approximately, 1 bark = 3 partitions */
    if (i > 8) return m1+m2;
    return (m1+m2)*table2[i];
  }

  if (m<15) {
    if (m > 0) {
      FLOAT8 f=1.0,r;
      if (i > 24) return m1+m2;
      if (i > 13) f = 1; else f = table3[i];
      r = (m-0)/15;
      return (m1+m2)*(table1[i]*r+f*(1-r));
    }
    if (i > 13) return m1+m2;
    return (m1+m2)*table3[i];
  }

  if (i > 24) return m1+m2;
  return (m1+m2)*table1[i];
}

/* 
 *   The spreading function.  Values returned in units of energy
 */
FLOAT8 s3_func(FLOAT8 bark) {
    
    FLOAT8 tempx,x,tempy,temp;
    tempx = bark;
    if (tempx>=0) tempx *= 3;
    else tempx *=1.5; 
    
    if(tempx>=0.5 && tempx<=2.5)
	{
	    temp = tempx - 0.5;
	    x = 8.0 * (temp*temp - 2.0 * temp);
	}
    else x = 0.0;
    tempx += 0.474;
    tempy = 15.811389 + 7.5*tempx - 17.5*sqrt(1.0+tempx*tempx);
    
    if (tempy <= -60.0) return  0.0;

    tempx = exp( (x + tempy)*LN_TO_LOG10 ); 

    /* Normalization.  The spreading function should be normalized so that:
         +inf
           /
           |  s3 [ bark ]  d(bark)   =  1
           /
         -inf
    */
    tempx /= .6609193;
    return tempx;
    
}

int
L3para_read(FLOAT8 sfreq, int *numlines_l,int *numlines_s, 
	FLOAT8 *minval,
	FLOAT8 s3_l[CBANDS][CBANDS], FLOAT8 s3_s[CBANDS][CBANDS],
	FLOAT8 *SNR, 
	int *bu_l, int *bo_l, FLOAT8 *w1_l, FLOAT8 *w2_l, 
	int *bu_s, int *bo_s, FLOAT8 *w1_s, FLOAT8 *w2_s,
	int *npart_l_orig,int *npart_l,int *npart_s_orig,int *npart_s)
{
  FLOAT8 bval_l[CBANDS], bval_s[CBANDS];
  FLOAT8 bval_l_width[CBANDS], bval_s_width[CBANDS];
/*  int   cbmax=0; */
  int  i,j;
/*  int freq_scale=1; */
  int partition[HBLKSIZE]; 



  /* compute numlines, the number of spectral lines in each partition band */
  /* each partition band should be about DELBARK wide. */
  j=0;
  for(i=0;i<CBANDS;i++)
    {
      FLOAT8 ji, bark1,bark2;
      int k,j2;

      j2 = j;
      j2 = Min(j2,BLKSIZE/2);
      
      do {
	/* find smallest j2 >= j so that  (bark - bark_l[i-1]) < DELBARK */
	ji = j;
	bark1 = freq2bark(sfreq*ji/BLKSIZE);
	
	++j2;
	ji = j2;
	bark2  = freq2bark(sfreq*ji/BLKSIZE);
      } while ((bark2 - bark1) < DELBARK  && j2<=BLKSIZE/2);

      for (k=j; k<j2; ++k)
	partition[k]=i;
      numlines_l[i]=(j2-j);
      j = j2;
      if (j > BLKSIZE/2) break;
    }
  *npart_l_orig = i+1;
  assert(*npart_l_orig <= CBANDS);

  /* compute which partition bands are in which scalefactor bands */
  { int i1,i2,sfb,start,end;
    FLOAT8 freq1,freq2;
    for ( sfb = 0; sfb < SBMAX_l; sfb++ ) {
      start = RO.scalefac_band.l[ sfb ];
      end   = RO.scalefac_band.l[ sfb+1 ];
      freq1 = sfreq*(start-.5)/(2*576);
      freq2 = sfreq*(end-1+.5)/(2*576);
		     
      i1 = floor(.5 + BLKSIZE*freq1/sfreq);
      if (i1<0) i1=0;
      i2 = floor(.5 + BLKSIZE*freq2/sfreq);
      if (i2>BLKSIZE/2) i2=BLKSIZE/2;

      //      RO.printf("longblock:  old: (%i,%i)  new: (%i,%i) %i %i \n",bu_l[sfb],bo_l[sfb],partition[i1],partition[i2],i1,i2);

      w1_l[sfb]=.5;
      w2_l[sfb]=.5;
      bu_l[sfb]=partition[i1];
      bo_l[sfb]=partition[i2];

    }
  }


  /* compute bark value and ATH of each critical band */
  j = 0;
  for ( i = 0; i < *npart_l_orig; i++ ) {
      int     k;
      FLOAT8  bark1,bark2;
      /* FLOAT8 mval,freq; */

      // Calculating the medium bark scaled frequency of the spectral lines
      // from j ... j + numlines[i]-1  (=spectral lines in parition band i)

      k         = numlines_l[i] - 1;
      bark1 = freq2bark(sfreq*(j+0)/BLKSIZE);
      bark2 = freq2bark(sfreq*(j+k)/BLKSIZE);
      bval_l[i] = .5*(bark1+bark2);

      bark1 = freq2bark(sfreq*(j+0-.5)/BLKSIZE);
      bark2 = freq2bark(sfreq*(j+k+.5)/BLKSIZE);
      bval_l_width[i] = bark2-bark1;

      RO.ATH.cb [i] = 1.e37; // preinit for minimum search
      for (k=0; k < numlines_l[i]; k++, j++) {
	FLOAT8  freq = sfreq*j/(1000.0*BLKSIZE);
	FLOAT8  level;
	assert( freq <= 24 );              // or only '<'
	//	freq = Min(.1,freq);       // ATH below 100 Hz constant, not further climbing
	level  = ATHformula (freq*1000) - 20;   // scale to FFT units; returned value is in dB
	level  = pow ( 10., 0.1*level );   // convert from dB -> energy
	level *= numlines_l [i];
	if ( level < RO.ATH.cb [i] )
	    RO.ATH.cb [i] = level;
      }


    }

  /* MINVAL.  For low freq, the strength of the masking is limited by minval
   * this is an ISO MPEG1 thing, dont know if it is really needed */
  for(i=0;i<*npart_l_orig;i++){
    double x = (-20+bval_l[i]*20.0/10.0);
    if (bval_l[i]>10) x = 0;
    minval[i]=pow(10.0,x/10);
  }







  /************************************************************************/
  /* SHORT BLOCKS */
  /************************************************************************/

  /* compute numlines */
  j=0;
  for(i=0;i<CBANDS;i++)
    {
      FLOAT8 ji, bark1,bark2;
      int k,j2;

      j2 = j;
      j2 = Min(j2,BLKSIZE_s/2);
      
      do {
	/* find smallest j2 >= j so that  (bark - bark_s[i-1]) < DELBARK */
	ji = j;
	bark1  = freq2bark(sfreq*ji/BLKSIZE_s);
	
	++j2;
	ji = j2;
	bark2  = freq2bark(sfreq*ji/BLKSIZE_s);

      } while ((bark2 - bark1) < DELBARK  && j2<=BLKSIZE_s/2);

      for (k=j; k<j2; ++k)
	partition[k]=i;
      numlines_s[i]=(j2-j);
      j = j2;
      if (j > BLKSIZE_s/2) break;
    }
  *npart_s_orig = i+1;
  assert(*npart_s_orig <= CBANDS);

  /* compute which partition bands are in which scalefactor bands */
  { int i1,i2,sfb,start,end;
    FLOAT8 freq1,freq2;
    for ( sfb = 0; sfb < SBMAX_s; sfb++ ) {
      start = RO.scalefac_band.s[ sfb ];
      end   = RO.scalefac_band.s[ sfb+1 ];
      freq1 = sfreq*(start-.5)/(2*192);
      freq2 = sfreq*(end-1+.5)/(2*192);
		     
      i1 = floor(.5 + BLKSIZE_s*freq1/sfreq);
      if (i1<0) i1=0;
      i2 = floor(.5 + BLKSIZE_s*freq2/sfreq);
      if (i2>BLKSIZE_s/2) i2=BLKSIZE_s/2;

      //RO.printf("shortblock: old: (%i,%i)  new: (%i,%i) %i %i \n",bu_s[sfb],bo_s[sfb], partition[i1],partition[i2],i1,i2);

      w1_s[sfb]=.5;
      w2_s[sfb]=.5;
      bu_s[sfb]=partition[i1];
      bo_s[sfb]=partition[i2];

    }
  }





  /* compute bark values of each critical band */
  j = 0;
  for(i=0;i<*npart_s_orig;i++)
    {
      int     k;
      FLOAT8  bark1,bark2,snr;
      k    = numlines_s[i] - 1;

      bark1 = freq2bark (sfreq*(j+0)/BLKSIZE_s);
      bark2 = freq2bark (sfreq*(j+k)/BLKSIZE_s); 
      bval_s[i] = .5*(bark1+bark2);

      bark1 = freq2bark (sfreq*(j+0-.5)/BLKSIZE_s);
      bark2 = freq2bark (sfreq*(j+k+.5)/BLKSIZE_s); 
      bval_s_width[i] = bark2-bark1;
      j        += k+1;
      
      /* SNR formula */
      if (bval_s[i]<13)
          snr=-8.25;
      else 
	  snr  = -4.5 * (bval_s[i]-13)/(24.0-13.0)  + 
	      -8.25*(bval_s[i]-24)/(13.0-24.0);

      SNR[i]=pow(10.0,snr/10.0);
    }






  /************************************************************************
   * Now compute the spreading function, s[j][i], the value of the spread-*
   * ing function, centered at band j, for band i, store for later use    *
   ************************************************************************/
  /* i.e.: sum over j to spread into signal barkval=i  
     NOTE: i and j are used opposite as in the ISO docs */
  for(i=0;i<*npart_l_orig;i++)    {
      for(j=0;j<*npart_l_orig;j++) 	{
  	  s3_l[i][j]=s3_func(bval_l[i]-bval_l[j])*bval_l_width[j];
      }
  }
  for(i=0;i<*npart_s_orig;i++)     {
      for(j=0;j<*npart_s_orig;j++) 	{
  	  s3_s[i][j]=s3_func(bval_s[i]-bval_s[j])*bval_s_width[j];
      }
  }
  



  /* compute: */
  /* npart_l_orig   = number of partition bands before convolution */
  /* npart_l  = number of partition bands after convolution */
  
  *npart_l=bo_l[NBPSY_l-1]+1;
  *npart_s=bo_s[NBPSY_s-1]+1;
  
  assert(*npart_l <= *npart_l_orig);
  assert(*npart_s <= *npart_s_orig);


    /* setup stereo demasking thresholds */
    /* formula reverse enginerred from plot in paper */
    for ( i = 0; i < NBPSY_s; i++ ) {
      FLOAT8 arg,mld;
      arg = freq2bark(sfreq*RO.scalefac_band.s[i]/(2*192));
      arg = (Min(arg, 15.5)/15.5);

      mld = 1.25*(1-cos(PI*arg))-2.5;
      RO.mld_s[i] = pow(10.0,mld);
    }
    for ( i = 0; i < NBPSY_l; i++ ) {
      FLOAT8 arg,mld;
      arg = freq2bark(sfreq*RO.scalefac_band.l[i]/(2*576));
      arg = (Min(arg, 15.5)/15.5);

      mld = 1.25*(1-cos(PI*arg))-2.5;
      RO.mld_l[i] = pow(10.0,mld);
    }

#define temporalmask_sustain_sec 0.01

    /* setup temporal masking */
    RO.decay = exp(-1.0*LOG10/(temporalmask_sustain_sec*sfreq/192.0));

    return 0;
}

int
psymodel_init(void)
{
    int i,j,sb,samplerate;
    FLOAT cwlimit;

    samplerate = RO.out_samplerate;
    RW.blocktype_old[0]=STOP_TYPE;
    RW.blocktype_old[1]=STOP_TYPE;
    RW.blocktype_old[0]=SHORT_TYPE;
    RW.blocktype_old[1]=SHORT_TYPE;

    for (i=0; i<4; ++i) {
      for (j=0; j<CBANDS; ++j) {
	RW.nb_1[i][j]=1e20;
	RW.nb_2[i][j]=1e20;
      }
      for ( sb = 0; sb < NBPSY_l; sb++ ) {
	RW.en[i].l[sb] = 1e20;
	RW.thm[i].l[sb] = 1e20;
      }
      for (j=0; j<3; ++j) {
	for ( sb = 0; sb < NBPSY_s; sb++ ) {
	  RW.en[i].s[sb][j] = 1e20;
	  RW.thm[i].s[sb][j] = 1e20;
	}
      }
    }

    RO.cw_lower_index=6;
    cwlimit=(FLOAT)8871.7;
    RO.cw_upper_index = cwlimit*1024.0/RO.out_samplerate;
    RO.cw_upper_index=Min(HBLKSIZE-4,RO.cw_upper_index);      /* j+3 < HBLKSIZE-1 */
    RO.cw_upper_index=Max(6,RO.cw_upper_index);

    for ( j = 0; j < HBLKSIZE; j++ )
      RW.cw[j] = 0.4f;

    i = L3para_read((FLOAT8) samplerate,RO.numlines_l,RO.numlines_s,
          RO.minval,RO.s3_l,RO.s3_s,RO.SNR_s,RO.bu_l,
          RO.bo_l,RO.w1_l,RO.w2_l, RO.bu_s,RO.bo_s,
          RO.w1_s,RO.w2_s,&RO.npart_l_orig,&RO.npart_l,
          &RO.npart_s_orig,&RO.npart_s );
    if (i!=0) return -1;

    /* npart_l_orig   = number of partition bands before convolution */
    /* npart_l  = number of partition bands after convolution */
    
    for (i=0; i<RO.npart_l; i++) {
      for (j = 0; j < RO.npart_l_orig; j++) {
	if (RO.s3_l[i][j] != 0.0)
	  break;
      }
      RO.s3ind[i][0] = j;
      
      for (j = RO.npart_l_orig - 1; j > 0; j--) {
	if (RO.s3_l[i][j] != 0.0)
	  break;
      }
      RO.s3ind[i][1] = j;
    }


    for (i=0; i<RO.npart_s; i++) {
      for (j = 0; j < RO.npart_s_orig; j++) {
	if (RO.s3_s[i][j] != 0.0)
	  break;
      }
      RO.s3ind_s[i][0] = j;
      
      for (j = RO.npart_s_orig - 1; j > 0; j--) {
	if (RO.s3_s[i][j] != 0.0)
	  break;
      }
      RO.s3ind_s[i][1] = j;
    }
    
    return 0;
}






